# future
from __future__ import annotations

# stdlib
from collections import OrderedDict
from collections.abc import Callable
import inspect
from inspect import Parameter
from inspect import signature
import types
from typing import Any
from typing import TYPE_CHECKING
from typing import _GenericAlias
from typing import cast
from typing import get_args
from typing import get_origin

# third party
from nacl.exceptions import BadSignatureError
from pydantic import BaseModel
from pydantic import ConfigDict
from pydantic import EmailStr
from pydantic import TypeAdapter
from result import OkErr
from result import Result
from typeguard import TypeCheckError
from typeguard import check_type

# relative
from ..abstract_server import AbstractServer
from ..protocol.data_protocol import PROTOCOL_TYPE
from ..protocol.data_protocol import get_data_protocol
from ..protocol.data_protocol import migrate_args_and_kwargs
from ..serde.deserialize import _deserialize
from ..serde.serializable import serializable
from ..serde.serialize import _serialize
from ..serde.signature import Signature
from ..serde.signature import signature_remove_context
from ..serde.signature import signature_remove_self
from ..server.credentials import SyftSigningKey
from ..server.credentials import SyftVerifyKey
from ..service.context import AuthedServiceContext
from ..service.context import ChangeContext
from ..service.metadata.server_metadata import ServerMetadataJSON
from ..service.response import SyftAttributeError
from ..service.response import SyftError
from ..service.response import SyftSuccess
from ..service.service import UserLibConfigRegistry
from ..service.service import UserServiceConfigRegistry
from ..service.user.user_roles import ServiceRole
from ..service.warnings import APIEndpointWarning
from ..service.warnings import WarningContext
from ..types.cache_object import CachedSyftObject
from ..types.identity import Identity
from ..types.syft_object import SYFT_OBJECT_VERSION_1
from ..types.syft_object import SyftBaseObject
from ..types.syft_object import SyftMigrationRegistry
from ..types.syft_object import SyftObject
from ..types.uid import LineageID
from ..types.uid import UID
from ..util.autoreload import autoreload_enabled
from ..util.markdown import as_markdown_python_code
from ..util.notebook_ui.components.tabulator_template import build_tabulator_table
from ..util.telemetry import instrument
from ..util.util import index_syft_by_module_name
from ..util.util import prompt_warning_message
from .connection import ServerConnection

if TYPE_CHECKING:
    # relative
    from ..server import Server
    from ..service.job.job_stash import Job


IPYNB_BACKGROUND_METHODS = {
    "getdoc",
    "_partialmethod",
    "__name__",
    "__code__",
    "__wrapped__",
    "__custom_documentations__",
    "__signature__",
    "__defaults__",
    "__kwdefaults__",
}

IPYNB_BACKGROUND_PREFIXES = ["_ipy", "_repr", "__ipython", "__pydantic"]


def _has_config_dict(t: Any) -> bool:
    return (
        # Use this instead of `issubclass`` to be compatible with python 3.10
        # `inspect.isclass(t) and issubclass(t, BaseModel)`` wouldn't work with
        # generics, e.g. `set[sy.UID]`, in python 3.10
        (hasattr(t, "__mro__") and BaseModel in t.__mro__)
        or hasattr(t, "__pydantic_config__")
    )


class APIRegistry:
    __api_registry__: dict[tuple, SyftAPI] = OrderedDict()

    @classmethod
    def set_api_for(
        cls,
        server_uid: UID | str,
        user_verify_key: SyftVerifyKey | str,
        api: SyftAPI,
    ) -> None:
        if isinstance(server_uid, str):
            server_uid = UID.from_string(server_uid)

        if isinstance(user_verify_key, str):
            user_verify_key = SyftVerifyKey.from_string(user_verify_key)

        key = (server_uid, user_verify_key)

        cls.__api_registry__[key] = api

    @classmethod
    def api_for(cls, server_uid: UID, user_verify_key: SyftVerifyKey) -> SyftAPI | None:
        key = (server_uid, user_verify_key)
        return cls.__api_registry__.get(key, None)

    @classmethod
    def get_all_api(cls) -> list[SyftAPI]:
        return list(cls.__api_registry__.values())

    @classmethod
    def get_by_recent_server_uid(cls, server_uid: UID) -> SyftAPI | None:
        for key, api in reversed(cls.__api_registry__.items()):
            if key[0] == server_uid:
                return api
        return None


@serializable()
class APIEndpoint(SyftObject):
    __canonical_name__ = "APIEndpoint"
    __version__ = SYFT_OBJECT_VERSION_1

    id: UID
    service_path: str
    module_path: str
    name: str
    description: str
    doc_string: str | None = None
    signature: Signature
    has_self: bool = False
    pre_kwargs: dict[str, Any] | None = None
    warning: APIEndpointWarning | None = None


@serializable()
class LibEndpoint(SyftBaseObject):
    __canonical_name__ = "LibEndpoint"
    __version__ = SYFT_OBJECT_VERSION_1

    # TODO: bad name, change
    service_path: str
    module_path: str
    name: str
    description: str
    doc_string: str | None = None
    signature: Signature
    has_self: bool = False
    pre_kwargs: dict[str, Any] | None = None


@serializable(attrs=["signature", "credentials", "serialized_message"])
class SignedSyftAPICall(SyftObject):
    __canonical_name__ = "SignedSyftAPICall"
    __version__ = SYFT_OBJECT_VERSION_1

    credentials: SyftVerifyKey
    signature: bytes
    serialized_message: bytes
    cached_deseralized_message: SyftAPICall | None = None

    @property
    def message(self) -> SyftAPICall:
        # from deserialize we might not have this attr because __init__ is skipped
        if not hasattr(self, "cached_deseralized_message"):
            self.cached_deseralized_message = None

        if self.cached_deseralized_message is None:
            self.cached_deseralized_message = _deserialize(
                blob=self.serialized_message, from_bytes=True
            )

        return self.cached_deseralized_message

    @property
    def is_valid(self) -> Result[SyftSuccess, SyftError]:
        try:
            _ = self.credentials.verify_key.verify(
                self.serialized_message, self.signature
            )
        except BadSignatureError:
            return SyftError(message="BadSignatureError")

        return SyftSuccess(message="Credentials are valid")


@instrument
@serializable()
class SyftAPICall(SyftObject):
    # version
    __canonical_name__ = "SyftAPICall"
    __version__ = SYFT_OBJECT_VERSION_1

    # fields
    server_uid: UID
    path: str
    args: list
    kwargs: dict[str, Any]
    blocking: bool = True

    def sign(self, credentials: SyftSigningKey) -> SignedSyftAPICall:
        signed_message = credentials.signing_key.sign(_serialize(self, to_bytes=True))

        return SignedSyftAPICall(
            credentials=credentials.verify_key,
            serialized_message=signed_message.message,
            signature=signed_message.signature,
        )

    def __repr__(self) -> str:
        return f"SyftAPICall(path={self.path}, args={self.args}, kwargs={self.kwargs}, blocking={self.blocking})"


@instrument
@serializable()
class SyftAPIData(SyftBaseObject):
    # version
    __canonical_name__ = "SyftAPIData"
    __version__ = SYFT_OBJECT_VERSION_1

    # fields
    data: Any = None

    def sign(self, credentials: SyftSigningKey) -> SignedSyftAPICall:
        signed_message = credentials.signing_key.sign(_serialize(self, to_bytes=True))

        return SignedSyftAPICall(
            credentials=credentials.verify_key,
            serialized_message=signed_message.message,
            signature=signed_message.signature,
        )


class RemoteFunction(SyftObject):
    __canonical_name__ = "RemoteFunction"
    __version__ = SYFT_OBJECT_VERSION_1
    __repr_attrs__ = [
        "id",
        "server_uid",
        "signature",
        "path",
    ]

    server_uid: UID
    signature: Signature
    refresh_api_callback: Callable | None = None
    path: str
    make_call: Callable
    pre_kwargs: dict[str, Any] | None = None
    communication_protocol: PROTOCOL_TYPE
    warning: APIEndpointWarning | None = None
    custom_function: bool = False

    @property
    def __ipython_inspector_signature_override__(self) -> Signature | None:
        return self.signature

    def prepare_args_and_kwargs(
        self, args: list | tuple, kwargs: dict[str, Any]
    ) -> SyftError | tuple[tuple, dict[str, Any]]:
        # Validate and migrate args and kwargs
        res = validate_callable_args_and_kwargs(args, kwargs, self.signature)
        if isinstance(res, SyftError):
            return res
        args, kwargs = res

        args, kwargs = migrate_args_and_kwargs(
            to_protocol=self.communication_protocol, args=args, kwargs=kwargs
        )

        return tuple(args), kwargs

    def function_call(
        self, path: str, *args: Any, cache_result: bool = True, **kwargs: Any
    ) -> Any:
        if "blocking" in self.signature.parameters:
            raise Exception(
                f"Signature {self.signature} can't have 'blocking' kwarg because it's reserved"
            )

        blocking = True
        if "blocking" in kwargs:
            blocking = bool(kwargs["blocking"])
            del kwargs["blocking"]

        res = self.prepare_args_and_kwargs(args, kwargs)
        if isinstance(res, SyftError):
            return res

        _valid_args, _valid_kwargs = res
        if self.pre_kwargs:
            _valid_kwargs.update(self.pre_kwargs)

        _valid_kwargs["communication_protocol"] = self.communication_protocol

        api_call = SyftAPICall(
            server_uid=self.server_uid,
            path=path,
            args=list(_valid_args),
            kwargs=_valid_kwargs,
            blocking=blocking,
        )

        allowed = self.warning.show() if self.warning else True
        if not allowed:
            return
        result = self.make_call(api_call=api_call, cache_result=cache_result)

        # TODO: annotate this on the service method decorator
        API_CALLS_THAT_REQUIRE_REFRESH = ["settings.enable_eager_execution"]

        if path in API_CALLS_THAT_REQUIRE_REFRESH:
            if self.refresh_api_callback is not None:
                self.refresh_api_callback()

        result, _ = migrate_args_and_kwargs(
            [result], kwargs={}, to_latest_protocol=True
        )
        result = result[0]
        return result

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        return self.function_call(self.path, *args, **kwargs)

    @property
    def mock(self) -> Any:
        if self.custom_function:
            remote_func = self

            class PrivateCustomAPIReference:
                def __call__(self, *args: Any, **kwargs: Any) -> Any:
                    return remote_func.function_call(
                        "api.call_public_in_jobs", *args, **kwargs
                    )

                @property
                def context(self) -> Any:
                    return remote_func.function_call("api.get_public_context")

            return PrivateCustomAPIReference()
        return SyftError(
            message="This function doesn't support mock/private calls as it's not custom."
        )

    @property
    def private(self) -> Any:
        if self.custom_function:
            remote_func = self

            class PrivateCustomAPIReference:
                def __call__(self, *args: Any, **kwargs: Any) -> Any:
                    return remote_func.function_call(
                        "api.call_private_in_jobs", *args, **kwargs
                    )

                @property
                def context(self) -> Any:
                    return remote_func.function_call("api.get_private_context")

            return PrivateCustomAPIReference()
        return SyftError(
            message="This function doesn't support mock/private calls as it's not custom."
        )

    def custom_function_actionobject_id(self) -> UID | SyftError:
        if self.custom_function and self.pre_kwargs is not None:
            custom_path = self.pre_kwargs.get("path", "")
            api_call = SyftAPICall(
                server_uid=self.server_uid,
                path="api.view",
                args=[custom_path],
                kwargs={},
            )
            endpoint = self.make_call(api_call=api_call)
            if isinstance(endpoint, SyftError):
                return endpoint
            return endpoint.action_object_id
        return SyftError(message="This function is not a custom function")

    def _repr_markdown_(self, wrap_as_python: bool = False, indent: int = 0) -> str:
        if self.custom_function and self.pre_kwargs is not None:
            custom_path = self.pre_kwargs.get("path", "")
            api_call = SyftAPICall(
                server_uid=self.server_uid,
                path="api.view",
                args=[custom_path],
                kwargs={},
            )
            endpoint = self.make_call(api_call=api_call)
            if isinstance(endpoint, SyftError):
                return endpoint._repr_html_()

            str_repr = "## API: " + custom_path + "\n"
            if endpoint.description is not None:
                text = endpoint.description.text
            else:
                text = ""
            str_repr += (
                "### Description: "
                + f'<span style="font-weight: lighter;">{text}</span><br>'
                + "\n"
            )
            str_repr += "#### Private Code:\n"
            not_accessible_code = "N / A"
            private_code_repr = endpoint.private_function or not_accessible_code
            public_code_repr = endpoint.mock_function or not_accessible_code
            str_repr += as_markdown_python_code(private_code_repr) + "\n"
            if endpoint.private_helper_functions:
                str_repr += "##### Helper Functions:\n"
                for helper_function in endpoint.private_helper_functions:
                    str_repr += as_markdown_python_code(helper_function) + "\n"
            str_repr += "#### Public Code:\n"
            str_repr += as_markdown_python_code(public_code_repr) + "\n"
            if endpoint.mock_helper_functions:
                str_repr += "##### Helper Functions:\n"
                for helper_function in endpoint.mock_helper_functions:
                    str_repr += as_markdown_python_code(helper_function) + "\n"
            return str_repr
        return super()._repr_markdown_()


class RemoteUserCodeFunction(RemoteFunction):
    __canonical_name__ = "RemoteUserFunction"
    __version__ = SYFT_OBJECT_VERSION_1
    __repr_attrs__ = RemoteFunction.__repr_attrs__ + ["user_code_id"]

    api: SyftAPI

    def prepare_args_and_kwargs(
        self, args: list | tuple, kwargs: dict[str, Any]
    ) -> tuple[tuple, dict[str, Any]] | SyftError:
        # relative
        from ..service.action.action_object import convert_to_pointers

        # Validate and migrate args and kwargs
        res = validate_callable_args_and_kwargs(args, kwargs, self.signature)
        if isinstance(res, SyftError):
            return res
        args, kwargs = res

        # Check remote function type to avoid function/method serialization
        # We can recover the function/method pointer by its UID in server side.
        for i in range(len(args)):
            if isinstance(args[i], RemoteFunction) and args[i].custom_function:
                args[i] = args[i].custom_function_id()

        for k, v in kwargs.items():
            if isinstance(v, RemoteFunction) and v.custom_function:
                kwargs[k] = v.custom_function_actionobject_id()

        args, kwargs = convert_to_pointers(
            api=self.api,
            server_uid=self.server_uid,
            args=args,
            kwargs=kwargs,
        )

        args, kwargs = migrate_args_and_kwargs(
            to_protocol=self.communication_protocol, args=args, kwargs=kwargs
        )

        return tuple(args), kwargs

    @property
    def user_code_id(self) -> UID | None:
        if self.pre_kwargs:
            return self.pre_kwargs.get("uid", None)
        else:
            return None

    @property
    def jobs(self) -> list[Job] | SyftError:
        if self.user_code_id is None:
            return SyftError(message="Could not find user_code_id")
        api_call = SyftAPICall(
            server_uid=self.server_uid,
            path="job.get_by_user_code_id",
            args=[self.user_code_id],
            kwargs={},
            blocking=True,
        )
        return self.make_call(api_call=api_call)


def generate_remote_function(
    api: SyftAPI,
    server_uid: UID,
    signature: Signature,
    path: str,
    make_call: Callable,
    pre_kwargs: dict[str, Any] | None,
    communication_protocol: PROTOCOL_TYPE,
    warning: APIEndpointWarning | None,
) -> RemoteFunction:
    if "blocking" in signature.parameters:
        raise Exception(
            f"Signature {signature} can't have 'blocking' kwarg because it's reserved"
        )

    # UserCodes are always code.call with a user_code_id
    if path == "code.call" and pre_kwargs is not None and "uid" in pre_kwargs:
        remote_function = RemoteUserCodeFunction(
            api=api,
            server_uid=server_uid,
            signature=signature,
            path=path,
            make_call=make_call,
            pre_kwargs=pre_kwargs,
            communication_protocol=communication_protocol,
            warning=warning,
            user_code_id=pre_kwargs["uid"],
        )
    else:
        custom_function = bool(path == "api.call_in_jobs")
        remote_function = RemoteFunction(
            server_uid=server_uid,
            refresh_api_callback=api.refresh_api_callback,
            signature=signature,
            path=path,
            make_call=make_call,
            pre_kwargs=pre_kwargs,
            communication_protocol=communication_protocol,
            warning=warning,
            custom_function=custom_function,
        )

    return remote_function


def generate_remote_lib_function(
    api: SyftAPI,
    server_uid: UID,
    signature: Signature,
    path: str,
    module_path: str,
    make_call: Callable,
    communication_protocol: PROTOCOL_TYPE,
    pre_kwargs: dict[str, Any],
) -> Any:
    if "blocking" in signature.parameters:
        raise Exception(
            f"Signature {signature} can't have 'blocking' kwarg because its reserved"
        )

    def wrapper(*args: Any, **kwargs: Any) -> SyftError | Any:
        # relative
        from ..service.action.action_object import TraceResultRegistry

        trace_result = TraceResultRegistry.get_trace_result_for_thread()

        if trace_result is not None:
            wrapper_make_call = trace_result.client.api.make_call  # type: ignore
            wrapper_server_uid = trace_result.client.api.server_uid  # type: ignore
        else:
            # somehow this is necessary to prevent shadowing problems
            wrapper_make_call = make_call
            wrapper_server_uid = server_uid
        blocking = True
        if "blocking" in kwargs:
            blocking = bool(kwargs["blocking"])
            del kwargs["blocking"]

        res = validate_callable_args_and_kwargs(args, kwargs, signature)

        if isinstance(res, SyftError):
            return res
        _valid_args, _valid_kwargs = res

        if pre_kwargs:
            _valid_kwargs.update(pre_kwargs)

        # relative
        from ..service.action.action_object import Action
        from ..service.action.action_object import ActionType
        from ..service.action.action_object import convert_to_pointers

        action_args, action_kwargs = convert_to_pointers(
            api, wrapper_server_uid, _valid_args, _valid_kwargs
        )

        # e.g. numpy.array -> numpy, array
        module, op = module_path.rsplit(".", 1)
        action = Action(
            path=module,
            op=op,
            remote_self=None,
            args=[x.syft_lineage_id for x in action_args],
            kwargs={k: v.syft_lineage_id for k, v in action_kwargs},
            action_type=ActionType.FUNCTION,
            # TODO: fix
            result_id=LineageID(UID(), 1),
        )
        service_args = [action]
        # TODO: implement properly
        if trace_result is not None:
            trace_result.result += [action]

        api_call = SyftAPICall(
            server_uid=wrapper_server_uid,
            path=path,
            args=service_args,
            kwargs={},
            blocking=blocking,
        )

        result = wrapper_make_call(api_call=api_call)
        return result

    wrapper.__ipython_inspector_signature_override__ = signature
    return wrapper


class APISubModulesView(SyftObject):
    __canonical_name__ = "APISubModulesView"
    __version__ = SYFT_OBJECT_VERSION_1

    submodule: str = ""
    endpoints: list[str] = []

    __syft_include_id_coll_repr__ = False

    def _coll_repr_(self) -> dict[str, Any]:
        return {"submodule": self.submodule, "endpoints": "\n".join(self.endpoints)}


@serializable(canonical_name="APIModule", version=1)
class APIModule:
    _modules: list[str]
    path: str
    refresh_callback: Callable | None

    def __init__(self, path: str, refresh_callback: Callable | None) -> None:
        self._modules = []
        self.path = path
        self.refresh_callback = refresh_callback

    def __dir__(self) -> list[str]:
        return self._modules + ["path"]

    def has_submodule(self, name: str) -> bool:
        """We use this as hasattr() triggers __getattribute__ which triggers recursion"""
        try:
            _ = object.__getattribute__(self, name)
            return True
        except AttributeError:
            return False

    def _add_submodule(
        self, attr_name: str, module_or_func: Callable | APIModule
    ) -> None:
        setattr(self, attr_name, module_or_func)
        self._modules.append(attr_name)

    def __getattr__(self, name: str) -> Any:
        try:
            return object.__getattribute__(self, name)
        except AttributeError:
            # if we fail, we refresh the api and try again
            # however, we dont want this to happen all the time because of ipy magic happening
            # in the background
            if (
                self.refresh_callback is not None
                and name not in IPYNB_BACKGROUND_METHODS
                and not any(
                    name.startswith(prefix) for prefix in IPYNB_BACKGROUND_PREFIXES
                )
            ):
                api = self.refresh_callback()
                try:
                    # get current path in the module tree
                    new_current_module = api.services
                    for submodule in self.path.split("."):
                        if submodule != "":
                            new_current_module = getattr(new_current_module, submodule)
                    # retry getting the attribute, if this fails, we throw an error
                    return object.__getattribute__(new_current_module, name)
                except AttributeError:
                    pass
            raise SyftAttributeError(
                f"'APIModule' api{self.path} object has no submodule or method '{name}', "
                "you may not have permission to access the module you are trying to access."
                "If you think this is an error, try calling `client.refresh()` to update the API."
            )

    def __getitem__(self, key: str | int) -> Any:
        if hasattr(self, "get_all"):
            return self.get_all()[key]
        raise NotImplementedError

    def _repr_html_(self) -> Any:
        if self.path == "settings":
            return self.get()._repr_html_()

        if not hasattr(self, "get_all"):

            def recursively_get_submodules(
                module: APIModule | Callable,
            ) -> list[APIModule | Callable]:
                children = [module]
                if isinstance(module, APIModule):
                    for submodule_name in module._modules:
                        submodule = getattr(module, submodule_name)
                        children += recursively_get_submodules(submodule)
                return children

            views = []
            for submodule_name in self._modules:
                submodule = getattr(self, submodule_name)
                children = recursively_get_submodules(submodule)
                child_paths = [
                    x.path for x in children if isinstance(x, RemoteFunction)
                ]
                views.append(
                    APISubModulesView(submodule=submodule_name, endpoints=child_paths)
                )

            return build_tabulator_table(views)

        # should never happen?
        results = self.get_all()
        return results._repr_html_()

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        return NotImplementedError


def debox_signed_syftapicall_response(
    signed_result: SignedSyftAPICall | Any,
) -> Any | SyftError:
    if not isinstance(signed_result, SignedSyftAPICall):
        return SyftError(message="The result is not signed")

    if not signed_result.is_valid:
        return SyftError(message="The result signature is invalid")
    return signed_result.message.data


def downgrade_signature(signature: Signature, object_versions: dict) -> Signature:
    migrated_parameters = []
    for parameter in signature.parameters.values():
        annotation = unwrap_and_migrate_annotation(
            parameter.annotation, object_versions
        )
        migrated_parameter = Parameter(
            name=parameter.name,
            default=parameter.default,
            annotation=annotation,
            kind=parameter.kind,
        )
        migrated_parameters.append(migrated_parameter)

    migrated_return_annotation = unwrap_and_migrate_annotation(
        signature.return_annotation, object_versions
    )

    try:
        new_signature = Signature(
            parameters=migrated_parameters,
            return_annotation=migrated_return_annotation,
        )
    except Exception as e:
        raise e

    return new_signature


def unwrap_and_migrate_annotation(annotation: Any, object_versions: dict) -> Any:
    args = get_args(annotation)
    origin = get_origin(annotation)
    if len(args) == 0:
        if (
            isinstance(annotation, type)
            and issubclass(annotation, SyftBaseObject)
            and annotation.__canonical_name__ in object_versions
        ):
            downgrade_to_version = int(
                max(object_versions[annotation.__canonical_name__])
            )
            downgrade_klass_name = SyftMigrationRegistry.__migration_version_registry__[
                annotation.__canonical_name__
            ][downgrade_to_version]
            new_arg = index_syft_by_module_name(downgrade_klass_name)
            return new_arg
        else:
            return annotation

    migrated_annotations = []
    for arg in args:
        migrated_annotation = unwrap_and_migrate_annotation(arg, object_versions)
        migrated_annotations.append(migrated_annotation)

    migrated_annotations_tuple = tuple(migrated_annotations)

    if hasattr(annotation, "copy_with"):
        return annotation.copy_with(migrated_annotations_tuple)
    elif origin is not None:
        return origin[migrated_annotations_tuple]
    else:
        return migrated_annotation[0]


def result_needs_api_update(api_call_result: Any) -> bool:
    # relative
    from ..service.request.request import Request
    from ..service.request.request import UserCodeStatusChange

    if isinstance(api_call_result, Request) and any(
        isinstance(x, UserCodeStatusChange) for x in api_call_result.changes
    ):
        return True
    if isinstance(api_call_result, SyftSuccess) and api_call_result.require_api_update:
        return True
    return False


@instrument
@serializable(
    attrs=[
        "endpoints",
        "server_uid",
        "server_name",
        "lib_endpoints",
        "communication_protocol",
    ]
)
class SyftAPI(SyftObject):
    # version
    __canonical_name__ = "SyftAPI"
    __version__ = SYFT_OBJECT_VERSION_1

    # fields
    connection: ServerConnection | None = None
    server_uid: UID | None = None
    server_name: str | None = None
    endpoints: dict[str, APIEndpoint]
    lib_endpoints: dict[str, LibEndpoint] | None = None
    api_module: APIModule | None = None
    libs: APIModule | None = None
    signing_key: SyftSigningKey | None = None
    # serde / storage rules
    refresh_api_callback: Callable | None = None
    __user_role: ServiceRole = ServiceRole.NONE
    communication_protocol: PROTOCOL_TYPE
    metadata: ServerMetadataJSON | None = None

    # informs getattr does not have nasty side effects
    __syft_allow_autocomplete__ = ["services"]

    def __dir__(self) -> list[str]:
        modules = getattr(self.api_module, "_modules", [])
        return ["services"] + modules

    def __syft_dir__(self) -> list[str]:
        modules = getattr(self.api_module, "_modules", [])
        return ["services"] + modules

    def __getattr__(self, name: str) -> Any:
        try:
            return getattr(self.api_module, name)
        except Exception:
            raise SyftAttributeError(
                f"'SyftAPI' object has no submodule or method '{name}', "
                "you may not have permission to access the module you are trying to access."
                "If you think this is an error, try calling `client.refresh()` to update the API."
            )

    @staticmethod
    def for_user(
        server: AbstractServer,
        communication_protocol: PROTOCOL_TYPE,
        user_verify_key: SyftVerifyKey | None = None,
    ) -> SyftAPI:
        # relative
        from ..service.api.api_service import APIService

        # TODO: Maybe there is a possibility of merging ServiceConfig and APIEndpoint
        from ..service.code.user_code_service import UserCodeService

        # find user role by verify_key
        # TODO: we should probably not allow empty verify keys but instead make user always register
        role = server.get_role_for_credentials(user_verify_key)
        _user_service_config_registry = UserServiceConfigRegistry.from_role(role)
        _user_lib_config_registry = UserLibConfigRegistry.from_user(user_verify_key)
        endpoints: dict[str, APIEndpoint] = {}
        lib_endpoints: dict[str, LibEndpoint] = {}
        warning_context = WarningContext(
            server=server, role=role, credentials=user_verify_key
        )

        # If server uses a higher protocol version than client, then
        # signatures needs to be downgraded.
        if server.current_protocol == "dev" and communication_protocol != "dev":
            # We assume dev is the highest staged protocol
            signature_needs_downgrade = True
        else:
            signature_needs_downgrade = server.current_protocol != "dev" and int(
                server.current_protocol
            ) > int(communication_protocol)
        data_protocol = get_data_protocol()

        if signature_needs_downgrade:
            object_version_for_protocol = data_protocol.get_object_versions(
                communication_protocol
            )

        for (
            path,
            service_config,
        ) in _user_service_config_registry.get_registered_configs().items():
            if not service_config.is_from_lib:
                service_warning = service_config.warning
                if service_warning:
                    service_warning = service_warning.message_from(warning_context)
                    service_warning.enabled = server.enable_warnings

                signature = (
                    downgrade_signature(
                        signature=service_config.signature,
                        object_versions=object_version_for_protocol,
                    )
                    if signature_needs_downgrade
                    else service_config.signature
                )

                endpoint = APIEndpoint(
                    service_path=path,
                    module_path=path,
                    name=service_config.public_name,
                    description="",
                    doc_string=service_config.doc_string,
                    signature=signature,  # TODO: Migrate signature based on communication protocol
                    has_self=False,
                    warning=service_warning,
                )
                endpoints[path] = endpoint

        for (
            path,
            lib_config,
        ) in _user_lib_config_registry.get_registered_configs().items():
            endpoint = LibEndpoint(
                service_path="action.execute",
                module_path=path,
                name=lib_config.public_name,
                description="",
                doc_string=lib_config.doc_string,
                signature=lib_config.signature,
                has_self=False,
            )
            lib_endpoints[path] = endpoint

        # 🟡 TODO 35: fix root context
        context = AuthedServiceContext(server=server, credentials=user_verify_key)
        method = server.get_method_with_context(
            UserCodeService.get_all_for_user, context
        )
        code_items = method()

        for code_item in code_items:
            path = "code.call"
            unique_path = f"code.call_{code_item.service_func_name}"
            endpoint = APIEndpoint(
                service_path=path,
                module_path=path,
                name=code_item.service_func_name,
                description="",
                doc_string=f"Users custom func {code_item.service_func_name}",
                signature=code_item.signature,
                has_self=False,
                pre_kwargs={"uid": code_item.id},
            )
            endpoints[unique_path] = endpoint

        # get admin defined custom api endpoints
        method = server.get_method_with_context(APIService.get_endpoints, context)
        custom_endpoints = method()
        for custom_endpoint in custom_endpoints:
            pre_kwargs = {"path": custom_endpoint.path}
            service_path = "api.call_in_jobs"
            path = custom_endpoint.path
            api_end = custom_endpoint.path.split(".")[-1]
            endpoint = APIEndpoint(
                service_path=service_path,
                module_path=path,
                name=api_end,
                description="",
                doc_string="",
                signature=custom_endpoint.signature,
                has_self=False,
                pre_kwargs=pre_kwargs,
            )
            endpoints[path] = endpoint

        return SyftAPI(
            server_name=server.name,
            server_uid=server.id,
            endpoints=endpoints,
            lib_endpoints=lib_endpoints,
            __user_role=role,
            communication_protocol=communication_protocol,
        )

    @property
    def user_role(self) -> ServiceRole:
        return self.__user_role

    def make_call(self, api_call: SyftAPICall, cache_result: bool = True) -> Result:
        signed_call = api_call.sign(credentials=self.signing_key)
        if self.connection is not None:
            signed_result = self.connection.make_call(signed_call)
        else:
            return SyftError(message="API connection is None")

        result = debox_signed_syftapicall_response(signed_result=signed_result)

        if isinstance(result, CachedSyftObject):
            if result.error_msg is not None:
                if cache_result:
                    msg = "Loading results from cache."
                    if result.error_msg:
                        msg = f"{result.error_msg}. {msg}"
                    prompt_warning_message(
                        message=msg,
                    )
                else:
                    result = SyftError(message=result.error_msg)
            if cache_result:
                result = result.result

        if isinstance(result, OkErr):
            if result.is_ok():
                result = result.ok()
            else:
                result = result.err()
        # we update the api when we create objects that change it
        self.update_api(result)
        return result

    def update_api(self, api_call_result: Any) -> None:
        # TODO: hacky stuff with typing and imports to prevent circular imports
        if result_needs_api_update(api_call_result):
            if self.refresh_api_callback is not None:
                self.refresh_api_callback()

    def _add_route(
        self, api_module: APIModule, endpoint: APIEndpoint, endpoint_method: Callable
    ) -> None:
        """Recursively create a module path to the route endpoint."""

        _modules = endpoint.module_path.split(".")[:-1] + [endpoint.name]

        _self = api_module
        _last_module = _modules.pop()
        while _modules:
            module = _modules.pop(0)
            if not _self.has_submodule(module):
                submodule_path = (
                    f"{_self.path}.{module}" if _self.path != "" else module
                )
                _self._add_submodule(
                    module,
                    APIModule(
                        path=submodule_path, refresh_callback=self.refresh_api_callback
                    ),
                )
            _self = getattr(_self, module)
        _self._add_submodule(_last_module, endpoint_method)

    def generate_endpoints(self) -> None:
        def build_endpoint_tree(
            endpoints: dict[str, LibEndpoint], communication_protocol: PROTOCOL_TYPE
        ) -> APIModule:
            api_module = APIModule(path="", refresh_callback=self.refresh_api_callback)
            for v in endpoints.values():
                signature = v.signature
                if not v.has_self:
                    signature = signature_remove_self(signature)
                signature = signature_remove_context(signature)
                if isinstance(v, APIEndpoint):
                    endpoint_function = generate_remote_function(
                        self,
                        self.server_uid,
                        signature,
                        v.service_path,
                        self.make_call,
                        pre_kwargs=v.pre_kwargs,
                        warning=v.warning,
                        communication_protocol=communication_protocol,
                    )
                elif isinstance(v, LibEndpoint):
                    endpoint_function = generate_remote_lib_function(
                        self,
                        self.server_uid,
                        signature,
                        v.service_path,
                        v.module_path,
                        self.make_call,
                        pre_kwargs=v.pre_kwargs,
                        communication_protocol=communication_protocol,
                    )

                endpoint_function.__doc__ = v.doc_string
                self._add_route(api_module, v, endpoint_function)
            return api_module

        if self.lib_endpoints is not None:
            self.libs = build_endpoint_tree(
                self.lib_endpoints, self.communication_protocol
            )
        self.api_module = build_endpoint_tree(
            self.endpoints, self.communication_protocol
        )

    @property
    def services(self) -> APIModule:
        if self.api_module is None:
            self.generate_endpoints()
        return cast(APIModule, self.api_module)

    @property
    def lib(self) -> APIModule:
        if self.libs is None:
            self.generate_endpoints()
        return cast(APIModule, self.libs)

    def has_service(self, service_name: str) -> bool:
        return hasattr(self.services, service_name)

    def has_lib(self, lib_name: str) -> bool:
        return hasattr(self.lib, lib_name)

    def __repr__(self) -> str:
        modules = self.services
        _repr_str = "client.api.services\n"
        if modules is not None:
            for attr_name in modules._modules:
                module_or_func = getattr(modules, attr_name)
                module_path_str = f"client.api.services.{attr_name}"
                _repr_str += f"\n{module_path_str}\n\n"
                if hasattr(module_or_func, "_modules"):
                    for func_name in module_or_func._modules:
                        func = getattr(module_or_func, func_name)
                        sig = getattr(
                            func, "__ipython_inspector_signature_override__", ""
                        )
                        _repr_str += f"{module_path_str}.{func_name}{sig}\n\n"
        return _repr_str


# code from here:
# https://github.com/ipython/ipython/blob/339c0d510a1f3cb2158dd8c6e7f4ac89aa4c89d8/IPython/core/oinspect.py#L370
def _render_signature(obj_signature: Signature, obj_name: str) -> str:
    """
    This was mostly taken from inspect.Signature.__str__.
    Look there for the comments.
    The only change is to add linebreaks when this gets too long.
    """
    result = []
    pos_only = False
    kw_only = True
    for param in obj_signature.parameters.values():
        if param.kind == inspect._POSITIONAL_ONLY:
            pos_only = True
        elif pos_only:
            result.append("/")
            pos_only = False

        if param.kind == inspect._VAR_POSITIONAL:
            kw_only = False
        elif param.kind == inspect._KEYWORD_ONLY and kw_only:
            result.append("*")
            kw_only = False

        result.append(str(param))

    if pos_only:
        result.append("/")

    # add up name, parameters, braces (2), and commas
    if len(obj_name) + sum(len(r) + 2 for r in result) > 75:
        # This doesn’t fit behind “Signature: ” in an inspect window.
        rendered = "{}(\n{})".format(obj_name, "".join(f"    {r},\n" for r in result))
    else:
        rendered = "{}({})".format(obj_name, ", ".join(result))

    if obj_signature.return_annotation is not inspect._empty:
        anno = inspect.formatannotation(obj_signature.return_annotation)
        rendered += f" -> {anno}"

    return rendered


def _getdef(self: Any, obj: Any, oname: str = "") -> str | None:
    """Return the call signature for any callable object.
    If any exception is generated, None is returned instead and the
    exception is suppressed."""
    try:
        return _render_signature(signature(obj), oname)
    except:  # noqa: E722
        return None


def monkey_patch_getdef(self: Any, obj: Any, oname: str = "") -> str | None:
    try:
        if hasattr(obj, "__ipython_inspector_signature_override__"):
            return _render_signature(
                obj.__ipython_inspector_signature_override__, oname
            )
        return _getdef(self, obj, oname)
    except Exception:
        return None


# try to monkeypatch IPython
try:
    # third party
    from IPython.core.oinspect import Inspector

    if not hasattr(Inspector, "_getdef_bak"):
        Inspector._getdef_bak = Inspector._getdef
        Inspector._getdef = types.MethodType(monkey_patch_getdef, Inspector)
except Exception:
    pass  # nosec


@serializable(canonical_name="ServerIdentity", version=1)
class ServerIdentity(Identity):
    server_name: str

    @staticmethod
    def from_api(api: SyftAPI) -> ServerIdentity:
        # stores the name root verify key of the datasite server
        if api.connection is None:
            raise ValueError(
                "{api}'s connection is None. Can't get the server identity"
            )
        server_metadata = api.connection.get_server_metadata(api.signing_key)
        return ServerIdentity(
            server_name=server_metadata.name,
            server_id=api.server_uid,
            verify_key=SyftVerifyKey.from_string(server_metadata.verify_key),
        )

    @classmethod
    def from_change_context(cls, context: ChangeContext) -> ServerIdentity:
        if context.server is None:
            raise ValueError(f"{context}'s server is None")
        return cls(
            server_name=context.server.name,
            server_id=context.server.id,
            verify_key=context.server.signing_key.verify_key,
        )

    @classmethod
    def from_server(cls, server: Server) -> ServerIdentity:
        return cls(
            server_name=server.name,
            server_id=server.id,
            verify_key=server.signing_key.verify_key,
        )

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, ServerIdentity):
            return False
        return (
            self.server_name == other.server_name
            and self.verify_key == other.verify_key
            and self.server_id == other.server_id
        )

    def __hash__(self) -> int:
        return hash((self.server_name, self.verify_key))

    def __repr__(self) -> str:
        return f"ServerIdentity <name={self.server_name}, id={self.server_id.short()}, 🔑={str(self.verify_key)[0:8]}>"


def validate_callable_args_and_kwargs(
    args: list, kwargs: dict, signature: Signature
) -> tuple[list, dict] | SyftError:
    _valid_kwargs = {}
    if "kwargs" in signature.parameters:
        _valid_kwargs = kwargs
    else:
        for key, value in kwargs.items():
            if key not in signature.parameters:
                return SyftError(
                    message=f"""Invalid parameter: `{key}`. Valid Parameters: {list(signature.parameters)}"""
                )
            param = signature.parameters[key]
            if isinstance(param.annotation, str):
                # 🟡 TODO 21: make this work for weird string type situations
                # happens when from __future__ import annotations in a class file
                t = index_syft_by_module_name(param.annotation)
            else:
                t = param.annotation

            if t is not inspect.Parameter.empty:
                try:
                    config_kw = (
                        {"config": ConfigDict(arbitrary_types_allowed=True)}
                        if not _has_config_dict(t)
                        else {}
                    )

                    # TypeAdapter only accepts `config` arg if `t` does not
                    # already contain a ConfigDict
                    # i.e model_config in BaseModel and __pydantic_config__ in
                    # other types.
                    TypeAdapter(t, **config_kw).validate_python(value)
                except Exception:
                    _type_str = getattr(t, "__name__", str(t))
                    return SyftError(
                        message=f"`{key}` must be of type `{_type_str}` not `{type(value).__name__}`"
                    )

            _valid_kwargs[key] = value

    # signature.parameters is an OrderedDict, therefore,
    # its fair to assume that order of args
    # and the signature.parameters should always match
    _valid_args = []
    if "args" in signature.parameters:
        _valid_args = args
    else:
        for (param_key, param), arg in zip(signature.parameters.items(), args):
            if param_key in _valid_kwargs:
                continue
            t = param.annotation
            msg = None
            try:
                if t is not inspect.Parameter.empty:
                    if isinstance(t, _GenericAlias) and type(None) in t.__args__:
                        for v in t.__args__:
                            if issubclass(v, EmailStr):
                                v = str
                            check_type(arg, v)  # raises Exception
                            break  # only need one to match
                    else:
                        check_type(arg, t)  # raises Exception
            except TypeCheckError:
                t_arg = type(arg)
                if (
                    autoreload_enabled()
                    and t.__module__ == t_arg.__module__
                    and t.__name__ == t_arg.__name__
                ):
                    # ignore error when autoreload_enabled()
                    pass
                else:
                    _type_str = getattr(t, "__name__", str(t))
                    msg = f"Arg is `{arg}`. \nIt must be of type `{_type_str}`, not `{type(arg).__name__}`"

            if msg:
                return SyftError(message=msg)

            _valid_args.append(arg)

    return _valid_args, _valid_kwargs


RemoteFunction.model_rebuild(force=True)
RemoteUserCodeFunction.model_rebuild(force=True)



# stdlib
from binascii import hexlify
from collections import defaultdict
import itertools
import logging
import socketserver
import sys
import threading
from threading import Event
import time
from time import sleep
from typing import Any
from typing import cast

# third party
from pydantic import field_validator
from result import Result
import zmq
from zmq import Frame
from zmq import LINGER
from zmq.error import ContextTerminated

# relative
from ...serde.deserialize import _deserialize
from ...serde.serializable import serializable
from ...serde.serialize import _serialize as serialize
from ...server.credentials import SyftVerifyKey
from ...service.action.action_object import ActionObject
from ...service.context import AuthedServiceContext
from ...types.base import SyftBaseModel
from ...types.syft_object import SYFT_OBJECT_VERSION_1
from ...types.syft_object import SyftObject
from ...types.uid import UID
from ...util.util import get_queue_address
from ..response import SyftError
from ..response import SyftSuccess
from ..service import AbstractService
from ..worker.worker_pool import ConsumerState
from ..worker.worker_pool import SyftWorker
from ..worker.worker_stash import WorkerStash
from .base_queue import AbstractMessageHandler
from .base_queue import QueueClient
from .base_queue import QueueClientConfig
from .base_queue import QueueConfig
from .base_queue import QueueConsumer
from .base_queue import QueueProducer
from .queue_stash import ActionQueueItem
from .queue_stash import QueueStash
from .queue_stash import Status

# Producer/Consumer heartbeat interval (in seconds)
HEARTBEAT_INTERVAL_SEC = 2

# Thread join timeout (in seconds)
THREAD_TIMEOUT_SEC = 30

# Max duration (in ms) to wait for ZMQ poller to return
ZMQ_POLLER_TIMEOUT_MSEC = 1000

# Duration (in seconds) after which a worker without a heartbeat will be marked as expired
WORKER_TIMEOUT_SEC = 60

# Duration (in seconds) after which producer without a heartbeat will be marked as expired
PRODUCER_TIMEOUT_SEC = 60

# Lock for working on ZMQ socket
ZMQ_SOCKET_LOCK = threading.Lock()

logger = logging.getLogger(__name__)


class QueueMsgProtocol:
    W_WORKER = b"MDPW01"
    W_READY = b"0x01"
    W_REQUEST = b"0x02"
    W_REPLY = b"0x03"
    W_HEARTBEAT = b"0x04"
    W_DISCONNECT = b"0x05"


MAX_RECURSION_NESTED_ACTIONOBJECTS = 5


class Timeout:
    def __init__(self, offset_sec: float):
        self.__offset = float(offset_sec)
        self.__next_ts: float = 0.0

        self.reset()

    @property
    def next_ts(self) -> float:
        return self.__next_ts

    def reset(self) -> None:
        self.__next_ts = self.now() + self.__offset

    def has_expired(self) -> bool:
        return self.now() >= self.__next_ts

    @staticmethod
    def now() -> float:
        return time.time()


class Service:
    def __init__(self, name: str) -> None:
        self.name = name
        self.requests: list[bytes] = []
        self.waiting: list[Worker] = []  # List of waiting workers


class Worker(SyftBaseModel):
    address: bytes
    identity: bytes
    service: Service | None = None
    syft_worker_id: UID | None = None
    expiry_t: Timeout = Timeout(WORKER_TIMEOUT_SEC)

    @field_validator("syft_worker_id", mode="before")
    @classmethod
    def set_syft_worker_id(cls, v: Any) -> Any:
        if isinstance(v, str):
            return UID(v)
        return v

    def has_expired(self) -> bool:
        return self.expiry_t.has_expired()

    def get_expiry(self) -> float:
        return self.expiry_t.next_ts

    def reset_expiry(self) -> None:
        self.expiry_t.reset()

    def _syft_worker(
        self, stash: WorkerStash, credentials: SyftVerifyKey
    ) -> Result[SyftWorker | None, str]:
        return stash.get_by_uid(credentials=credentials, uid=self.syft_worker_id)

    def __str__(self) -> str:
        svc = self.service.name if self.service else None
        return (
            f"Worker(addr={self.address!r}, id={self.identity!r}, service={svc}, "
            f"syft_worker_id={self.syft_worker_id!r})"
        )


@serializable(canonical_name="ZMQProducer", version=1)
class ZMQProducer(QueueProducer):
    INTERNAL_SERVICE_PREFIX = b"mmi."

    def __init__(
        self,
        queue_name: str,
        queue_stash: QueueStash,
        worker_stash: WorkerStash,
        port: int,
        context: AuthedServiceContext,
    ) -> None:
        self.id = UID().short()
        self.port = port
        self.queue_stash = queue_stash
        self.worker_stash = worker_stash
        self.queue_name = queue_name
        self.auth_context = context
        self._stop = Event()
        self.post_init()

    @property
    def address(self) -> str:
        return get_queue_address(self.port)

    def post_init(self) -> None:
        """Initialize producer state."""

        self.services: dict[str, Service] = {}
        self.workers: dict[bytes, Worker] = {}
        self.waiting: list[Worker] = []
        self.heartbeat_t = Timeout(HEARTBEAT_INTERVAL_SEC)
        self.context = zmq.Context(1)
        self.socket = self.context.socket(zmq.ROUTER)
        self.socket.setsockopt(LINGER, 1)
        self.socket.setsockopt_string(zmq.IDENTITY, self.id)
        self.poll_workers = zmq.Poller()
        self.poll_workers.register(self.socket, zmq.POLLIN)
        self.bind(f"tcp://*:{self.port}")
        self.thread: threading.Thread | None = None
        self.producer_thread: threading.Thread | None = None

    def close(self) -> None:
        self._stop.set()
        try:
            if self.thread:
                self.thread.join(THREAD_TIMEOUT_SEC)
                if self.thread.is_alive():
                    logger.error(
                        f"ZMQProducer message sending thread join timed out during closing. "
                        f"Queue name {self.queue_name}, "
                    )
                self.thread = None

            if self.producer_thread:
                self.producer_thread.join(THREAD_TIMEOUT_SEC)
                if self.producer_thread.is_alive():
                    logger.error(
                        f"ZMQProducer queue thread join timed out during closing. "
                        f"Queue name {self.queue_name}, "
                    )
                self.producer_thread = None

            self.poll_workers.unregister(self.socket)
        except Exception as e:
            logger.exception("Failed to unregister poller.", exc_info=e)
        finally:
            self.socket.close()
            self.context.destroy()

            # self._stop.clear()

    @property
    def action_service(self) -> AbstractService:
        if self.auth_context.server is not None:
            return self.auth_context.server.get_service("ActionService")
        else:
            raise Exception(f"{self.auth_context} does not have a server.")

    def contains_unresolved_action_objects(self, arg: Any, recursion: int = 0) -> bool:
        """recursively check collections for unresolved action objects"""
        if isinstance(arg, UID):
            arg = self.action_service.get(self.auth_context, arg).ok()
            return self.contains_unresolved_action_objects(arg, recursion=recursion + 1)
        if isinstance(arg, ActionObject):
            if not arg.syft_resolved:
                res = self.action_service.get(self.auth_context, arg)
                if res.is_err():
                    return True
                arg = res.ok()
                if not arg.syft_resolved:
                    return True
            arg = arg.syft_action_data

        try:
            value = False
            if isinstance(arg, list):
                for elem in arg:
                    value = self.contains_unresolved_action_objects(
                        elem, recursion=recursion + 1
                    )
                    if value:
                        return True
            if isinstance(arg, dict):
                for elem in arg.values():
                    value = self.contains_unresolved_action_objects(
                        elem, recursion=recursion + 1
                    )
                    if value:
                        return True
            return value
        except Exception as e:
            logger.exception("Failed to resolve action objects.", exc_info=e)
            return True

    def read_items(self) -> None:
        while True:
            if self._stop.is_set():
                break
            try:
                sleep(1)

                # Items to be queued
                items_to_queue = self.queue_stash.get_by_status(
                    self.queue_stash.partition.root_verify_key,
                    status=Status.CREATED,
                ).ok()

                items_to_queue = [] if items_to_queue is None else items_to_queue

                # Queue Items that are in the processing state
                items_processing = self.queue_stash.get_by_status(
                    self.queue_stash.partition.root_verify_key,
                    status=Status.PROCESSING,
                ).ok()

                items_processing = [] if items_processing is None else items_processing

                for item in itertools.chain(items_to_queue, items_processing):
                    # TODO: if resolving fails, set queueitem to errored, and jobitem as well
                    if item.status == Status.CREATED:
                        if isinstance(item, ActionQueueItem):
                            action = item.kwargs["action"]
                            if self.contains_unresolved_action_objects(
                                action.args
                            ) or self.contains_unresolved_action_objects(action.kwargs):
                                continue

                        msg_bytes = serialize(item, to_bytes=True)
                        worker_pool = item.worker_pool.resolve_with_context(
                            self.auth_context
                        )
                        worker_pool = worker_pool.ok()
                        service_name = worker_pool.name
                        service: Service | None = self.services.get(service_name)

                        # Skip adding message if corresponding service/pool
                        # is not registered.
                        if service is None:
                            continue

                        # append request message to the corresponding service
                        # This list is processed in dispatch method.

                        # TODO: Logic to evaluate the CAN RUN Condition
                        service.requests.append(msg_bytes)
                        item.status = Status.PROCESSING
                        res = self.queue_stash.update(item.syft_client_verify_key, item)
                        if res.is_err():
                            logger.error(
                                f"Failed to update queue item={item} error={res.err()}"
                            )
                    elif item.status == Status.PROCESSING:
                        # Evaluate Retry condition here
                        # If job running and timeout or job status is KILL
                        # or heartbeat fails
                        # or container id doesn't exists, kill process or container
                        # else decrease retry count and mark status as CREATED.
                        pass
            except Exception as e:
                print(e, file=sys.stderr)
                item.status = Status.ERRORED
                res = self.queue_stash.update(item.syft_client_verify_key, item)
                if res.is_err():
                    logger.error(
                        f"Failed to update queue item={item} error={res.err()}"
                    )

    def run(self) -> None:
        self.thread = threading.Thread(target=self._run)
        self.thread.start()

        self.producer_thread = threading.Thread(target=self.read_items)
        self.producer_thread.start()

    def send(self, worker: bytes, message: bytes | list[bytes]) -> None:
        worker_obj = self.require_worker(worker)
        self.send_to_worker(worker_obj, QueueMsgProtocol.W_REQUEST, message)

    def bind(self, endpoint: str) -> None:
        """Bind producer to endpoint."""
        self.socket.bind(endpoint)
        logger.info(f"ZMQProducer endpoint: {endpoint}")

    def send_heartbeats(self) -> None:
        """Send heartbeats to idle workers if it's time"""
        if self.heartbeat_t.has_expired():
            for worker in self.waiting:
                self.send_to_worker(worker, QueueMsgProtocol.W_HEARTBEAT)
            self.heartbeat_t.reset()

    def purge_workers(self) -> None:
        """Look for & kill expired workers.

        Workers are oldest to most recent, so we stop at the first alive worker.
        """
        # work on a copy of the iterator
        for worker in self.waiting:
            res = worker._syft_worker(self.worker_stash, self.auth_context.credentials)
            if res.is_err() or (syft_worker := res.ok()) is None:
                logger.info(f"Failed to retrieve SyftWorker {worker.syft_worker_id}")
                continue

            if worker.has_expired() or syft_worker.to_be_deleted:
                logger.info(f"Deleting expired worker id={worker}")
                self.delete_worker(worker, syft_worker.to_be_deleted)

                # relative
                from ...service.worker.worker_service import WorkerService

                worker_service = cast(
                    WorkerService, self.auth_context.server.get_service(WorkerService)
                )
                worker_service._delete(self.auth_context, syft_worker)

    def update_consumer_state_for_worker(
        self, syft_worker_id: UID, consumer_state: ConsumerState
    ) -> None:
        if self.worker_stash is None:
            logger.error(  # type: ignore[unreachable]
                f"ZMQProducer worker stash not defined for {self.queue_name} - {self.id}"
            )
            return

        try:
            # Check if worker is present in the database
            worker = self.worker_stash.get_by_uid(
                credentials=self.worker_stash.partition.root_verify_key,
                uid=syft_worker_id,
            )
            if worker.is_ok() and worker.ok() is None:
                return

            res = self.worker_stash.update_consumer_state(
                credentials=self.worker_stash.partition.root_verify_key,
                worker_uid=syft_worker_id,
                consumer_state=consumer_state,
            )
            if res.is_err():
                logger.error(
                    f"Failed to update consumer state for worker id={syft_worker_id} "
                    f"to state: {consumer_state} error={res.err()}",
                )
        except Exception as e:
            logger.error(
                f"Failed to update consumer state for worker id: {syft_worker_id} to state {consumer_state}",
                exc_info=e,
            )

    def worker_waiting(self, worker: Worker) -> None:
        """This worker is now waiting for work."""
        # Queue to broker and service waiting lists
        if worker not in self.waiting:
            self.waiting.append(worker)
        if worker.service is not None and worker not in worker.service.waiting:
            worker.service.waiting.append(worker)
        worker.reset_expiry()
        self.update_consumer_state_for_worker(worker.syft_worker_id, ConsumerState.IDLE)
        self.dispatch(worker.service, None)

    def dispatch(self, service: Service, msg: bytes) -> None:
        """Dispatch requests to waiting workers as possible"""
        if msg is not None:  # Queue message if any
            service.requests.append(msg)

        self.purge_workers()
        while service.waiting and service.requests:
            # One worker consuming only one message at a time.
            msg = service.requests.pop(0)
            worker = service.waiting.pop(0)
            self.waiting.remove(worker)
            self.send_to_worker(worker, QueueMsgProtocol.W_REQUEST, msg)

    def send_to_worker(
        self,
        worker: Worker,
        command: bytes,
        msg: bytes | list | None = None,
    ) -> None:
        """Send message to worker.

        If message is provided, sends that message.
        """

        if self.socket.closed:
            logger.warning("Socket is closed. Cannot send message.")
            return

        if msg is None:
            msg = []
        elif not isinstance(msg, list):
            msg = [msg]

        # ZMQProducer send frames: [address, empty, header, command, ...data]
        core = [worker.address, b"", QueueMsgProtocol.W_WORKER, command]
        msg = core + msg

        if command != QueueMsgProtocol.W_HEARTBEAT:
            # log everything except the last frame which contains serialized data
            logger.info(f"ZMQProducer send: {core}")

        with ZMQ_SOCKET_LOCK:
            try:
                self.socket.send_multipart(msg)
            except zmq.ZMQError as e:
                logger.error("ZMQProducer send error", exc_info=e)

    def _run(self) -> None:
        try:
            while True:
                if self._stop.is_set():
                    logger.info("ZMQProducer thread stopped")
                    return

                for service in self.services.values():
                    self.dispatch(service, None)

                items = None

                try:
                    items = self.poll_workers.poll(ZMQ_POLLER_TIMEOUT_MSEC)
                except Exception as e:
                    logger.exception("ZMQProducer poll error", exc_info=e)

                if items:
                    msg = self.socket.recv_multipart()

                    if len(msg) < 3:
                        logger.error(f"ZMQProducer invalid recv: {msg}")
                        continue

                    # ZMQProducer recv frames: [address, empty, header, command, ...data]
                    (address, _, header, command, *data) = msg

                    if command != QueueMsgProtocol.W_HEARTBEAT:
                        # log everything except the last frame which contains serialized data
                        logger.info(f"ZMQProducer recv: {msg[:4]}")

                    if header == QueueMsgProtocol.W_WORKER:
                        self.process_worker(address, command, data)
                    else:
                        logger.error(f"Invalid message header: {header}")

                self.send_heartbeats()
                self.purge_workers()
        except Exception as e:
            logger.exception("ZMQProducer thread exception", exc_info=e)

    def require_worker(self, address: bytes) -> Worker:
        """Finds the worker (creates if necessary)."""
        identity = hexlify(address)
        worker = self.workers.get(identity)
        if worker is None:
            worker = Worker(identity=identity, address=address)
            self.workers[identity] = worker
        return worker

    def process_worker(self, address: bytes, command: bytes, data: list[bytes]) -> None:
        worker_ready = hexlify(address) in self.workers
        worker = self.require_worker(address)

        if QueueMsgProtocol.W_READY == command:
            service_name = data.pop(0).decode()
            syft_worker_id = data.pop(0).decode()
            if worker_ready:
                # Not first command in session or Reserved service name
                # If worker was already present, then we disconnect it first
                # and wait for it to re-register itself to the producer. This ensures that
                # we always have a healthy worker in place that can talk to the producer.
                self.delete_worker(worker, True)
            else:
                # Attach worker to service and mark as idle
                if service_name in self.services:
                    service: Service | None = self.services.get(service_name)
                else:
                    service = Service(service_name)
                    self.services[service_name] = service
                if service is not None:
                    worker.service = service
                logger.info(f"New worker: {worker}")
                worker.syft_worker_id = UID(syft_worker_id)
                self.worker_waiting(worker)

        elif QueueMsgProtocol.W_HEARTBEAT == command:
            if worker_ready:
                # If worker is ready then reset expiry
                # and add it to worker waiting list
                # if not already present
                self.worker_waiting(worker)
            else:
                logger.info(f"Got heartbeat, but worker not ready. {worker}")
                self.delete_worker(worker, True)
        elif QueueMsgProtocol.W_DISCONNECT == command:
            logger.info(f"Removing disconnected worker: {worker}")
            self.delete_worker(worker, False)
        else:
            logger.error(f"Invalid command: {command!r}")

    def delete_worker(self, worker: Worker, disconnect: bool) -> None:
        """Deletes worker from all data structures, and deletes worker."""
        if disconnect:
            self.send_to_worker(worker, QueueMsgProtocol.W_DISCONNECT)

        if worker.service and worker in worker.service.waiting:
            worker.service.waiting.remove(worker)

        if worker in self.waiting:
            self.waiting.remove(worker)

        self.workers.pop(worker.identity, None)

        if worker.syft_worker_id is not None:
            self.update_consumer_state_for_worker(
                worker.syft_worker_id, ConsumerState.DETACHED
            )

    @property
    def alive(self) -> bool:
        return not self.socket.closed


@serializable(attrs=["_subscriber"], canonical_name="ZMQConsumer", version=1)
class ZMQConsumer(QueueConsumer):
    def __init__(
        self,
        message_handler: AbstractMessageHandler,
        address: str,
        queue_name: str,
        service_name: str,
        syft_worker_id: UID | None = None,
        worker_stash: WorkerStash | None = None,
        verbose: bool = False,
    ) -> None:
        self.address = address
        self.message_handler = message_handler
        self.service_name = service_name
        self.queue_name = queue_name
        self.context = zmq.Context()
        self.poller = zmq.Poller()
        self.socket = None
        self.verbose = verbose
        self.id = UID().short()
        self._stop = Event()
        self.syft_worker_id = syft_worker_id
        self.worker_stash = worker_stash
        self.post_init()

    def reconnect_to_producer(self) -> None:
        """Connect or reconnect to producer"""
        if self.socket:
            self.poller.unregister(self.socket)  # type: ignore[unreachable]
            self.socket.close()
        self.socket = self.context.socket(zmq.DEALER)
        self.socket.linger = 0
        self.socket.setsockopt_string(zmq.IDENTITY, self.id)
        self.socket.connect(self.address)
        self.poller.register(self.socket, zmq.POLLIN)

        logger.info(f"Connecting Worker id={self.id} to broker addr={self.address}")

        # Register queue with the producer
        self.send_to_producer(
            QueueMsgProtocol.W_READY,
            [self.service_name.encode(), str(self.syft_worker_id).encode()],
        )

    def post_init(self) -> None:
        self.thread: threading.Thread | None = None
        self.heartbeat_t = Timeout(HEARTBEAT_INTERVAL_SEC)
        self.producer_ping_t = Timeout(PRODUCER_TIMEOUT_SEC)
        self.reconnect_to_producer()

    def disconnect_from_producer(self) -> None:
        self.send_to_producer(QueueMsgProtocol.W_DISCONNECT)

    def close(self) -> None:
        self.disconnect_from_producer()
        self._stop.set()
        try:
            if self.thread is not None:
                self.thread.join(timeout=THREAD_TIMEOUT_SEC)
                if self.thread.is_alive():
                    logger.error(
                        f"ZMQConsumer thread join timed out during closing. "
                        f"SyftWorker id {self.syft_worker_id}, "
                        f"service name {self.service_name}."
                    )
                self.thread = None
            self.poller.unregister(self.socket)
        except Exception as e:
            logger.error("Failed to unregister worker.", exc_info=e)
        finally:
            self.socket.close()
            self.context.destroy()
            # self._stop.clear()

    def send_to_producer(
        self,
        command: bytes,
        msg: bytes | list | None = None,
    ) -> None:
        """Send message to producer.

        If no msg is provided, creates one internally
        """
        if self.socket.closed:
            logger.warning("Socket is closed. Cannot send message.")
            return

        if msg is None:
            msg = []
        elif not isinstance(msg, list):
            msg = [msg]

        # ZMQConsumer send frames: [empty, header, command, ...data]
        core = [b"", QueueMsgProtocol.W_WORKER, command]
        msg = core + msg

        if command != QueueMsgProtocol.W_HEARTBEAT:
            logger.info(f"ZMQ Consumer send: {core}")

        with ZMQ_SOCKET_LOCK:
            try:
                self.socket.send_multipart(msg)
            except zmq.ZMQError as e:
                logger.error("ZMQConsumer send error", exc_info=e)

    def _run(self) -> None:
        """Send reply, if any, to producer and wait for next request."""
        try:
            while True:
                if self._stop.is_set():
                    logger.info("ZMQConsumer thread stopped")
                    return

                try:
                    items = self.poller.poll(ZMQ_POLLER_TIMEOUT_MSEC)
                except ContextTerminated:
                    logger.info("Context terminated")
                    return
                except Exception as e:
                    logger.error("ZMQ poll error", exc_info=e)
                    continue

                if items:
                    msg = self.socket.recv_multipart()

                    # mark as alive
                    self.set_producer_alive()

                    if len(msg) < 3:
                        logger.error(f"ZMQConsumer invalid recv: {msg}")
                        continue

                    # Message frames recieved by consumer:
                    # [empty, header, command, ...data]
                    (_, _, command, *data) = msg

                    if command != QueueMsgProtocol.W_HEARTBEAT:
                        # log everything except the last frame which contains serialized data
                        logger.info(f"ZMQConsumer recv: {msg[:-4]}")

                    if command == QueueMsgProtocol.W_REQUEST:
                        # Call Message Handler
                        try:
                            message = data.pop()
                            self.associate_job(message)
                            self.message_handler.handle_message(
                                message=message,
                                syft_worker_id=self.syft_worker_id,
                            )
                        except Exception as e:
                            logger.exception("Couldn't handle message", exc_info=e)
                        finally:
                            self.clear_job()
                    elif command == QueueMsgProtocol.W_HEARTBEAT:
                        self.set_producer_alive()
                    elif command == QueueMsgProtocol.W_DISCONNECT:
                        self.reconnect_to_producer()
                    else:
                        logger.error(f"ZMQConsumer invalid command: {command}")
                else:
                    if not self.is_producer_alive():
                        logger.info("Producer check-alive timed out. Reconnecting.")
                        self.reconnect_to_producer()
                        self.set_producer_alive()

                if not self._stop.is_set():
                    self.send_heartbeat()

        except zmq.ZMQError as e:
            if e.errno == zmq.ETERM:
                logger.info("zmq.ETERM")
            else:
                logger.exception("zmq.ZMQError", exc_info=e)
        except Exception as e:
            logger.exception("ZMQConsumer thread exception", exc_info=e)

    def set_producer_alive(self) -> None:
        self.producer_ping_t.reset()

    def is_producer_alive(self) -> bool:
        # producer timer is within timeout
        return not self.producer_ping_t.has_expired()

    def send_heartbeat(self) -> None:
        if self.heartbeat_t.has_expired() and self.is_producer_alive():
            self.send_to_producer(QueueMsgProtocol.W_HEARTBEAT)
            self.heartbeat_t.reset()

    def run(self) -> None:
        self.thread = threading.Thread(target=self._run)
        self.thread.start()

    def associate_job(self, message: Frame) -> None:
        try:
            queue_item = _deserialize(message, from_bytes=True)
            self._set_worker_job(queue_item.job_id)
        except Exception as e:
            logger.exception("Could not associate job", exc_info=e)

    def clear_job(self) -> None:
        self._set_worker_job(None)

    def _set_worker_job(self, job_id: UID | None) -> None:
        if self.worker_stash is not None:
            consumer_state = (
                ConsumerState.IDLE if job_id is None else ConsumerState.CONSUMING
            )
            res = self.worker_stash.update_consumer_state(
                credentials=self.worker_stash.partition.root_verify_key,
                worker_uid=self.syft_worker_id,
                consumer_state=consumer_state,
            )
            if res.is_err():
                logger.error(
                    f"Failed to update consumer state for {self.service_name}-{self.id}, error={res.err()}"
                )

    @property
    def alive(self) -> bool:
        return not self.socket.closed and self.is_producer_alive()


@serializable()
class ZMQClientConfig(SyftObject, QueueClientConfig):
    __canonical_name__ = "ZMQClientConfig"
    __version__ = SYFT_OBJECT_VERSION_1

    id: UID | None = None  # type: ignore[assignment]
    hostname: str = "127.0.0.1"
    queue_port: int | None = None
    # TODO: setting this to false until we can fix the ZMQ
    # port issue causing tests to randomly fail
    create_producer: bool = False
    n_consumers: int = 0
    consumer_service: str | None = None


@serializable(attrs=["host"], canonical_name="ZMQClient", version=1)
class ZMQClient(QueueClient):
    """ZMQ Client for creating producers and consumers."""

    producers: dict[str, ZMQProducer]
    consumers: defaultdict[str, list[ZMQConsumer]]

    def __init__(self, config: ZMQClientConfig) -> None:
        self.host = config.hostname
        self.producers = {}
        self.consumers = defaultdict(list)
        self.config = config

    @staticmethod
    def _get_free_tcp_port(host: str) -> int:
        with socketserver.TCPServer((host, 0), None) as s:
            free_port = s.server_address[1]

        return free_port

    def add_producer(
        self,
        queue_name: str,
        port: int | None = None,
        queue_stash: QueueStash | None = None,
        worker_stash: WorkerStash | None = None,
        context: AuthedServiceContext | None = None,
    ) -> ZMQProducer:
        """Add a producer of a queue.

        A queue can have at most one producer attached to it.
        """

        if port is None:
            if self.config.queue_port is None:
                self.config.queue_port = self._get_free_tcp_port(self.host)
                port = self.config.queue_port
            else:
                port = self.config.queue_port

        producer = ZMQProducer(
            queue_name=queue_name,
            queue_stash=queue_stash,
            port=port,
            context=context,
            worker_stash=worker_stash,
        )
        self.producers[queue_name] = producer
        return producer

    def add_consumer(
        self,
        queue_name: str,
        message_handler: AbstractMessageHandler,
        service_name: str,
        address: str | None = None,
        worker_stash: WorkerStash | None = None,
        syft_worker_id: UID | None = None,
    ) -> ZMQConsumer:
        """Add a consumer to a queue

        A queue should have at least one producer attached to the group.

        """

        if address is None:
            address = get_queue_address(self.config.queue_port)

        consumer = ZMQConsumer(
            queue_name=queue_name,
            message_handler=message_handler,
            address=address,
            service_name=service_name,
            syft_worker_id=syft_worker_id,
            worker_stash=worker_stash,
        )
        self.consumers[queue_name].append(consumer)

        return consumer

    def send_message(
        self,
        message: bytes,
        queue_name: str,
        worker: bytes | None = None,
    ) -> SyftSuccess | SyftError:
        producer = self.producers.get(queue_name)
        if producer is None:
            return SyftError(
                message=f"No producer attached for queue: {queue_name}. Please add a producer for it."
            )
        try:
            producer.send(message=message, worker=worker)
        except Exception as e:
            # stdlib
            return SyftError(
                message=f"Failed to send message to: {queue_name} with error: {e}"
            )
        return SyftSuccess(
            message=f"Successfully queued message to : {queue_name}",
        )

    def close(self) -> SyftError | SyftSuccess:
        try:
            for consumers in self.consumers.values():
                for consumer in consumers:
                    # make sure look is stopped
                    consumer.close()

            for producer in self.producers.values():
                # make sure loop is stopped
                producer.close()
                # close existing connection.
        except Exception as e:
            return SyftError(message=f"Failed to close connection: {e}")

        return SyftSuccess(message="All connections closed.")

    def purge_queue(self, queue_name: str) -> SyftError | SyftSuccess:
        if queue_name not in self.producers:
            return SyftError(message=f"No producer running for : {queue_name}")

        producer = self.producers[queue_name]

        # close existing connection.
        producer.close()

        # add a new connection
        self.add_producer(queue_name=queue_name, address=producer.address)  # type: ignore

        return SyftSuccess(message=f"Queue: {queue_name} successfully purged")

    def purge_all(self) -> SyftError | SyftSuccess:
        for queue_name in self.producers:
            self.purge_queue(queue_name=queue_name)

        return SyftSuccess(message="Successfully purged all queues.")


@serializable(canonical_name="ZMQQueueConfig", version=1)
class ZMQQueueConfig(QueueConfig):
    def __init__(
        self,
        client_type: type[ZMQClient] | None = None,
        client_config: ZMQClientConfig | None = None,
        thread_workers: bool = False,
    ):
        self.client_type = client_type or ZMQClient
        self.client_config: ZMQClientConfig = client_config or ZMQClientConfig()
        self.thread_workers = thread_workers



# stdlib
from collections import defaultdict
from collections.abc import Iterable
from collections.abc import Mapping
from collections.abc import Set
import logging
import re
from typing import Any

# relative
from .util import full_name_with_qualname
from .util import sanitize_html

TABLE_INDEX_KEY = "_table_repr_index"

logger = logging.getLogger(__name__)


def _syft_in_mro(self: Any, item: Any) -> bool:
    if hasattr(type(item), "mro") and type(item) != type:
        # if unbound method, supply self
        if hasattr(type(item).mro, "__self__"):
            mro = type(item).mro()
        else:
            mro = type(item).mro(type(item))  # type: ignore

    elif hasattr(item, "mro") and type(item) != type:
        mro = item.mro()
    else:
        mro = str(self)  # type: ignore

    return "syft" in str(mro).lower()


def _get_values_for_table_repr(obj: Any) -> Any:
    if isinstance(obj, Mapping):
        values = list(obj.values())
    elif isinstance(obj, Set):
        values = list(obj)
    else:
        values = obj

    return values


def _get_grid_template_columns(first_value: Any) -> tuple[str | None, str | None]:
    grid_template_cols = getattr(first_value, "__table_coll_widths__", None)
    if isinstance(grid_template_cols, list):
        grid_template_columns = " ".join(grid_template_cols)
        grid_template_cell_columns = "unset"
    else:
        grid_template_columns = None
        grid_template_cell_columns = None
    return grid_template_columns, grid_template_cell_columns


def _create_table_rows(
    _self: Mapping | Iterable,
    is_homogenous: bool,
    extra_fields: list | None = None,
    add_index: bool = True,
) -> list[dict[str, Any]]:
    """
    Creates row data for a table based on input object obj.

    If valid table data cannot be created, an empty list is returned.

    Args:
        _self (Mapping | Iterable): The input data as a Mapping or Iterable.
        is_homogenous (bool): A boolean indicating whether the data is homogenous.
        extra_fields (list | None, optional): Additional fields to include in the table. Defaults to None.
        add_index (bool, optional): Whether to add an index column. Defaults to True.

    Returns:
        list[dict[str, Any]]: A list of dictionaries where each dictionary represents a row in the table.

    """

    if extra_fields is None:
        extra_fields = []

    cols = defaultdict(list)

    for item in iter(_self.items() if isinstance(_self, Mapping) else _self):
        # unpack dict
        if isinstance(_self, Mapping):
            key, item = item
            cols["key"].append(key)

        # get id
        id_ = getattr(item, "id", None)
        include_id = getattr(item, "__syft_include_id_coll_repr__", True)
        if id_ is not None and include_id:
            cols["id"].append({"value": str(id_), "type": "clipboard"})

        if type(item) == type:
            t = full_name_with_qualname(item)
        else:
            try:
                t = item.__class__.__name__
            except Exception:
                t = item.__repr__()

        if not is_homogenous:
            cols["type"].append(t)

        # if has _coll_repr_

        if hasattr(item, "_coll_repr_"):
            ret_val = item._coll_repr_()
            if "id" in ret_val:
                del ret_val["id"]
            for key in ret_val.keys():
                cols[key].append(ret_val[key])
        else:
            for field in extra_fields:
                value = item
                try:
                    attrs = field.split(".")
                    for i, attr in enumerate(attrs):
                        # find indexing like abc[1]
                        res = re.search(r"\[[+-]?\d+\]", attr)
                        has_index = False
                        if res:
                            has_index = True
                            index_str = res.group()
                            index = int(index_str.replace("[", "").replace("]", ""))
                            attr = attr.replace(index_str, "")

                        value = getattr(value, attr, None)
                        if isinstance(value, list) and has_index:
                            value = value[index]
                        # If the object has a special representation when nested we will use that instead
                        if (
                            hasattr(value, "__repr_syft_nested__")
                            and i == len(attrs) - 1
                        ):
                            value = value.__repr_syft_nested__()
                        if (
                            isinstance(value, list)
                            and i == len(attrs) - 1
                            and len(value) > 0
                            and hasattr(value[0], "__repr_syft_nested__")
                        ):
                            value = [
                                (
                                    x.__repr_syft_nested__()
                                    if hasattr(x, "__repr_syft_nested__")
                                    else x
                                )
                                for x in value
                            ]
                    if value is None:
                        value = "n/a"

                except Exception as e:
                    print(e)
                    value = None
                cols[field].append(sanitize_html(str(value)))

    col_lengths = {len(cols[col]) for col in cols.keys()}
    if len(col_lengths) != 1:
        logger.debug("Cannot create table for items with different number of fields.")
        return []

    num_rows = col_lengths.pop()
    if add_index and TABLE_INDEX_KEY not in cols:
        cols[TABLE_INDEX_KEY] = list(range(num_rows))

    # NOTE cannot use Pandas, not all values can be in a DataFrame (dict/list/...)
    rows = []
    for i in range(num_rows):
        row = {}
        for col in cols.keys():
            row[col] = cols[col][i]
        rows.append(row)

    return rows


def _sort_table_rows(rows: list[dict[str, Any]], sort_key: str) -> list[dict[str, Any]]:
    try:
        sort_values = [row[sort_key] for row in rows]
    except KeyError:
        # Not all rows have the sort_key, do not sort
        return rows

    # relative
    from ..types.datetime import DateTime
    from ..types.datetime import str_is_datetime

    if all(isinstance(v, str) and str_is_datetime(v) for v in sort_values):
        sort_values = [DateTime.from_str(v) for v in sort_values]

    reverse_sort = False
    if isinstance(sort_values[0], DateTime):
        sort_values = [d.utc_timestamp for d in sort_values]
        reverse_sort = True

    rows_sorted = [
        row
        for _, row in sorted(
            zip(sort_values, rows),
            reverse=reverse_sort,
            key=lambda pair: pair[0],
        )
    ]

    return rows_sorted


def prepare_table_data(
    obj: Any,
    add_index: bool = True,
) -> tuple[list[dict], dict]:
    """
    Creates table data and metadata for a given object.

    If a tabular representation cannot be created, an empty list and empty dict are returned instead.

    Args:
        obj (Any): The input object for which table data is prepared.
        add_index (bool, optional): Whether to add an index column to the table. Defaults to True.

    Returns:
        tuple: A tuple (table_data, table_metadata) where table_data is a list of dictionaries
        where each dictionary represents a row in the table and table_metadata is a dictionary
        containing metadata about the table such as name, icon, etc.

    """

    values = _get_values_for_table_repr(obj)
    if len(values) == 0:
        return [], {}

    # check first value and obj itself to see if syft in mro. If not, don't create table
    first_value = values[0]
    if not _syft_in_mro(obj, first_value):
        return [], {}

    extra_fields = getattr(first_value, "__repr_attrs__", [])
    is_homogenous = len({type(x) for x in values}) == 1
    if is_homogenous:
        sort_key = getattr(first_value, "__table_sort_attr__", None) or "created_at"
        cls_name = first_value.__class__.__name__
        grid_template_columns, grid_template_cell_columns = _get_grid_template_columns(
            first_value
        )
    else:
        sort_key = "created_at"
        cls_name = ""
        grid_template_columns = None
        grid_template_cell_columns = None

    table_data = _create_table_rows(
        _self=obj,
        is_homogenous=is_homogenous,
        extra_fields=extra_fields,
        add_index=add_index,
    )
    # if empty result, collection objects have no table representation
    if not table_data:
        return [], {}

    table_data = _sort_table_rows(table_data, sort_key)

    table_metadata = {
        "name": f"{cls_name} {obj.__class__.__name__.capitalize()}",
        "columns": list(table_data[0].keys()),
        "icon": getattr(first_value, "icon", None),
        "grid_template_columns": grid_template_columns,
        "grid_template_cell_columns": grid_template_cell_columns,
    }

    return table_data, table_metadata



# stdlib
import contextlib
import os
import signal
import subprocess  # nosec
import tempfile
import time


@contextlib.contextmanager
def pyspy() -> None:  # type: ignore
    """Profile a block of code using py-spy. Intended for development purposes only.

    Example:
    ```
    with pyspy():
        # do some work
        a = [i for i in range(1000000)]
    ```
    """
    fd, fname = tempfile.mkstemp(".svg")
    os.close(fd)

    command = [
        "sudo",
        "-S",
        "py-spy",
        "record",
        "-r",
        "100",
        "-o",
        fname,
        "--pid",
        str(os.getpid()),
    ]
    process = subprocess.Popen(command, preexec_fn=os.setsid)  # nosec

    start_time = time.time()
    yield process
    end_time = time.time()

    print(f"Execution time: {end_time - start_time}")
    try:
        os.killpg(os.getpgid(process.pid), signal.SIGINT)
        os.chmod(fname, 0o444)
    except Exception as e:
        print(f"Error: {e}")



# stdlib
from enum import Enum
from typing import Any

# relative
from ...serde.serializable import serializable
from ...server.credentials import SyftVerifyKey
from ...types.uid import UID


@serializable(canonical_name="ActionPermission", version=1)
class ActionPermission(Enum):
    OWNER = 1
    READ = 2
    ALL_READ = 4
    WRITE = 8
    ALL_WRITE = 32
    EXECUTE = 64
    ALL_EXECUTE = 128


COMPOUND_ACTION_PERMISSION = {
    ActionPermission.ALL_READ,
    ActionPermission.ALL_WRITE,
    ActionPermission.ALL_EXECUTE,
}


@serializable(canonical_name="ActionObjectPermission", version=1)
class ActionObjectPermission:
    def __init__(
        self,
        uid: UID,
        permission: ActionPermission,
        credentials: SyftVerifyKey | None = None,
    ):
        if credentials is None:
            if permission not in COMPOUND_ACTION_PERMISSION:
                raise Exception(f"{permission} not in {COMPOUND_ACTION_PERMISSION}")
        self.uid = uid
        self.credentials = credentials
        self.permission = permission

    @classmethod
    def from_permission_string(
        cls, uid: UID, permission_string: str
    ) -> "ActionObjectPermission":
        if permission_string.startswith("ALL_"):
            permission = ActionPermission[permission_string]
            verify_key = None
        else:
            verify_key_str, perm_str = permission_string.split("_", 1)
            permission = ActionPermission[perm_str]
            verify_key = SyftVerifyKey.from_string(verify_key_str)

        return cls(uid=uid, permission=permission, credentials=verify_key)

    @property
    def permission_string(self) -> str:
        if self.permission in COMPOUND_ACTION_PERMISSION:
            return f"{self.permission.name}"
        else:
            if self.credentials is not None:
                return f"{self.credentials.verify}_{self.permission.name}"
            return f"{self.permission.name}"

    def _coll_repr_(self) -> dict[str, Any]:
        return {
            "uid": str(self.uid),
            "credentials": str(self.credentials),
            "permission": str(self.permission),
        }

    def __repr__(self) -> str:
        if self.credentials is not None:
            return f"[{self.permission.name}: {self.uid} as {self.credentials.verify}]"
        else:
            # TODO: somehow, __repr__ is only triggered in this case.
            # Maybe fixed by change from <> brackets to [] so it now prints in htlm?
            return self.permission_string


class ActionObjectOWNER(ActionObjectPermission):
    def __init__(self, uid: UID, credentials: SyftVerifyKey):
        self.uid = uid
        self.credentials = credentials
        self.permission = ActionPermission.OWNER


class ActionObjectREAD(ActionObjectPermission):
    def __init__(self, uid: UID, credentials: SyftVerifyKey):
        self.uid = uid
        self.credentials = credentials
        self.permission = ActionPermission.READ


class ActionObjectWRITE(ActionObjectPermission):
    def __init__(self, uid: UID, credentials: SyftVerifyKey):
        self.uid = uid
        self.credentials = credentials
        self.permission = ActionPermission.WRITE


class ActionObjectEXECUTE(ActionObjectPermission):
    def __init__(self, uid: UID, credentials: SyftVerifyKey):
        self.uid = uid
        self.credentials = credentials
        self.permission = ActionPermission.EXECUTE


@serializable(canonical_name="StoragePermission", version=1)
class StoragePermission:
    def __init__(self, uid: UID, server_uid: UID):
        self.uid = uid
        self.server_uid = server_uid

    def __repr__(self) -> str:
        return f"StoragePermission: {self.uid} on {self.server_uid}"

    def _coll_repr_(self) -> dict[str, Any]:
        return {
            "uid": str(self.uid),
            "server_uid": str(self.server_uid),
        }



# stdlib

# third party
from result import Err
from result import Ok
from result import Result

# relative
from ...serde.serializable import serializable
from ...server.credentials import SyftVerifyKey
from ...store.document_store import BaseUIDStoreStash
from ...store.document_store import DocumentStore
from ...store.document_store import PartitionKey
from ...store.document_store import PartitionSettings
from ...store.document_store import QueryKeys
from ...types.uid import UID
from ...util.telemetry import instrument
from ..action.action_permissions import ActionObjectPermission
from ..action.action_permissions import ActionPermission
from .worker_pool import ConsumerState
from .worker_pool import SyftWorker

WorkerContainerNamePartitionKey = PartitionKey(key="container_name", type_=str)


@instrument
@serializable(canonical_name="WorkerStash", version=1)
class WorkerStash(BaseUIDStoreStash):
    object_type = SyftWorker
    settings: PartitionSettings = PartitionSettings(
        name=SyftWorker.__canonical_name__, object_type=SyftWorker
    )

    def __init__(self, store: DocumentStore) -> None:
        super().__init__(store=store)

    def set(
        self,
        credentials: SyftVerifyKey,
        obj: SyftWorker,
        add_permissions: list[ActionObjectPermission] | None = None,
        add_storage_permission: bool = True,
        ignore_duplicates: bool = False,
    ) -> Result[SyftWorker, str]:
        # By default all worker pools have all read permission
        add_permissions = [] if add_permissions is None else add_permissions
        add_permissions.append(
            ActionObjectPermission(uid=obj.id, permission=ActionPermission.ALL_READ)
        )
        return super().set(
            credentials,
            obj,
            add_permissions=add_permissions,
            ignore_duplicates=ignore_duplicates,
            add_storage_permission=add_storage_permission,
        )

    def get_worker_by_name(
        self, credentials: SyftVerifyKey, worker_name: str
    ) -> Result[SyftWorker | None, str]:
        qks = QueryKeys(qks=[WorkerContainerNamePartitionKey.with_obj(worker_name)])
        return self.query_one(credentials=credentials, qks=qks)

    def update_consumer_state(
        self, credentials: SyftVerifyKey, worker_uid: UID, consumer_state: ConsumerState
    ) -> Result[str, str]:
        res = self.get_by_uid(credentials=credentials, uid=worker_uid)
        if res.is_err():
            return Err(
                f"Failed to retrieve Worker with id: {worker_uid}. Error: {res.err()}"
            )
        worker: SyftWorker | None = res.ok()
        if worker is None:
            return Err(f"Worker with id: {worker_uid} not found")
        worker.consumer_state = consumer_state
        update_res = self.update(credentials=credentials, obj=worker)
        if update_res.is_err():
            return Err(
                f"Failed to update Worker with id: {worker_uid}. Error: {update_res.err()}"
            )
        return Ok(f"Successfully updated Worker with id: {worker_uid}")



# future
from __future__ import annotations

# stdlib
import secrets
from typing import Any
from typing import TYPE_CHECKING

# third party
from typing_extensions import Self

# relative
from ...abstract_server import AbstractServer
from ...client.client import HTTPConnection
from ...client.client import PythonConnection
from ...client.client import ServerConnection
from ...client.client import SyftClient
from ...serde.serializable import serializable
from ...server.worker_settings import WorkerSettings
from ...types.syft_object import SYFT_OBJECT_VERSION_1
from ...types.syft_object import SyftObject
from ...types.transforms import TransformContext
from ...types.uid import UID
from ..context import AuthedServiceContext
from ..context import ServerServiceContext
from ..response import SyftError

if TYPE_CHECKING:
    # relative
    from .server_peer import ServerPeer


@serializable(canonical_name="ServerRoute", version=1)
class ServerRoute:
    def client_with_context(
        self, context: ServerServiceContext
    ) -> SyftClient | SyftError:
        """
        Convert the current route (self) to a connection (either HTTP, Veilid or Python)
        and create a SyftClient from the connection.

        Args:
            context (ServerServiceContext): The ServerServiceContext containing the server information.

        Returns:
            SyftClient | SyftError: Returns the created SyftClient, or SyftError
                if the client type is not valid or if the context's server is None.
        """
        connection = route_to_connection(route=self, context=context)
        client_type = connection.get_client_type()
        if isinstance(client_type, SyftError):
            return client_type
        return client_type(
            connection=connection, credentials=context.server.signing_key
        )

    def validate_with_context(
        self, context: AuthedServiceContext
    ) -> ServerPeer | SyftError:
        # relative
        from .server_peer import ServerPeer

        # Step 1: Check if the given route is able to reach the given server
        # As we allow the user to give custom routes, we need to check the reachability of the route
        self_client = self.client_with_context(context=context)

        # generating a random challenge
        random_challenge = secrets.token_bytes(16)
        challenge_signature = self_client.api.services.network.ping(random_challenge)

        if isinstance(challenge_signature, SyftError):
            return challenge_signature

        try:
            # Verifying if the challenge is valid
            context.server.verify_key.verify_key.verify(
                random_challenge, challenge_signature
            )
        except Exception:
            return SyftError(message="Signature Verification Failed in ping")

        # Step 2: Create a Server Peer with the given route
        self_server_peer: ServerPeer = context.server.settings.to(ServerPeer)
        self_server_peer.server_routes.append(self)

        return self_server_peer


@serializable()
class HTTPServerRoute(SyftObject, ServerRoute):
    __canonical_name__ = "HTTPServerRoute"
    __version__ = SYFT_OBJECT_VERSION_1

    id: UID | None = None  # type: ignore
    host_or_ip: str
    private: bool = False
    protocol: str = "http"
    port: int = 80
    proxy_target_uid: UID | None = None
    priority: int = 1
    rtunnel_token: str | None = None

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, HTTPServerRoute):
            return False
        return hash(self) == hash(other)

    def __hash__(self) -> int:
        return (
            hash(self.host_or_ip)
            + hash(self.port)
            + hash(self.protocol)
            + hash(self.proxy_target_uid)
            + hash(self.rtunnel_token)
        )

    def __str__(self) -> str:
        return f"{self.protocol}://{self.host_or_ip}:{self.port}"


@serializable()
class PythonServerRoute(SyftObject, ServerRoute):
    __canonical_name__ = "PythonServerRoute"
    __version__ = SYFT_OBJECT_VERSION_1

    id: UID | None = None  # type: ignore
    worker_settings: WorkerSettings
    proxy_target_uid: UID | None = None
    priority: int = 1

    @property
    def server(self) -> AbstractServer | None:
        # relative
        from ...server.worker import Worker

        server = Worker(
            id=self.worker_settings.id,
            name=self.worker_settings.name,
            server_type=self.worker_settings.server_type,
            server_side_type=self.worker_settings.server_side_type,
            signing_key=self.worker_settings.signing_key,
            document_store_config=self.worker_settings.document_store_config,
            action_store_config=self.worker_settings.action_store_config,
            processes=1,
        )
        return server

    @classmethod
    def with_server(cls, server: AbstractServer) -> Self:
        worker_settings = WorkerSettings.from_server(server)
        return cls(id=worker_settings.id, worker_settings=worker_settings)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, PythonServerRoute):
            return False
        return hash(self) == hash(other)

    def __hash__(self) -> int:
        return (
            hash(self.worker_settings.id)
            + hash(self.worker_settings.name)
            + hash(self.worker_settings.server_type)
            + hash(self.worker_settings.server_side_type)
            + hash(self.worker_settings.signing_key)
        )

    def __str__(self) -> str:
        return "PythonServerRoute"


@serializable()
class VeilidServerRoute(SyftObject, ServerRoute):
    __canonical_name__ = "VeilidServerRoute"
    __version__ = SYFT_OBJECT_VERSION_1

    vld_key: str
    proxy_target_uid: UID | None = None
    priority: int = 1

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, VeilidServerRoute):
            return False
        return hash(self) == hash(other)

    def __hash__(self) -> int:
        return hash(self.vld_key) + hash(self.proxy_target_uid)


ServerRouteTypeV1 = HTTPServerRoute | PythonServerRoute | VeilidServerRoute
ServerRouteType = HTTPServerRoute | PythonServerRoute


def route_to_connection(
    route: ServerRoute, context: TransformContext | None = None
) -> ServerConnection:
    if isinstance(route, HTTPServerRoute):
        return route.to(HTTPConnection, context=context)
    elif isinstance(route, PythonServerRoute):
        return route.to(PythonConnection, context=context)
    else:
        raise ValueError(f"Route {route} is not supported.")


def connection_to_route(connection: ServerConnection) -> ServerRoute:
    if isinstance(connection, HTTPConnection):
        return connection.to(HTTPServerRoute)
    elif isinstance(connection, PythonConnection):  # type: ignore[unreachable]
        return connection.to(PythonServerRoute)
    else:
        raise ValueError(f"Connection {connection} is not supported.")
